/*
 * giftofgf.c -- Converts a GIF file to a FGF file
 *
 * Jeremy Wohl - January 7, 1995
 *
 * Portions of this file are derived from the giftopnm.c module by David
 * Koblas, as included in the NetPBM package (release 7 December 1993).
 *
 * The deflate routines were written by Mark Adler et al. and
 * originate in the zip package distributed by the Info-Zip group.
 *
 * The Graphics Interchange Format(c) is the Copyright property of
 * CompuServe Incorporated.  GIF(sm) is a Service Mark property of
 * CompuServe Incorporated.
 *
 * This software is provided "as is" without express or implied warranty.
 *
 */

/* +-------------------------------------------------------------------+ */
/* | Copyright 1990, 1991, 1993, David Koblas.  (koblas@netcom.com)    | */
/* |   Permission to use, copy, modify, and distribute this software   | */
/* |   and its documentation for any purpose and without fee is hereby | */
/* |   granted, provided that the above copyright notice appear in all | */
/* |   copies and that both that copyright notice and this permission  | */
/* |   notice appear in supporting documentation.  This software is    | */
/* |   provided "as is" without express or implied warranty.           | */
/* +-------------------------------------------------------------------+ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXCOLORMAPSIZE  256

#define TRUE    1
#define FALSE   0

#define CM_RED         0
#define CM_GREEN       1
#define CM_BLUE        2

#define MAX_LZW_BITS   12

#define INTERLACE      0x40
#define LOCALCOLORMAP  0x80
#define BitSet(byte, bit)      (((byte) & (bit)) == (bit))

#define ReadOK(file, buffer, len)   (fread(buffer, len, 1, file) != 0)
#define WriteOK(file, buffer, len) (fwrite(buffer, len, 1, file) != 0)

#define LM_to_uint(a,b)  (((b)<<8)|(a))

static struct {
       unsigned int    Width;
       unsigned int    Height;
       unsigned char   ColorMap[3][MAXCOLORMAPSIZE];
       unsigned int    BitPixel;
       unsigned int    ColorResolution;
       unsigned int    Background;
       unsigned int    AspectRatio;
       /*
       **
       */
       int             GrayScale;
} GifScreen;

static struct {
       int     transparent;
       int     delayTime;
       int     inputFlag;
       int     disposal;
} Gif89 = { -1, -1, -1, 0 };

static int ReadColorMap ( FILE *fd, int number, unsigned char buffer[3][MAXCOLORMAPSIZE]);
static int WriteColorMap (FILE * fd, int number, unsigned char buffer [3][MAXCOLORMAPSIZE]);
static int DoExtension (FILE *fd, int label);
static int GetDataBlock (FILE *fd, unsigned char  *buf);
static int GetAndWriteDataBlock (FILE * fd, unsigned char * buf);
static int GetCode (FILE *fd, int code_size, int flag);
static int LZWReadByte (FILE *fd, int flag, int input_code_size);
static void ConvertGIF (FILE * gif_file, FILE * fgf_file);
static void ConvertImageData (FILE * gif_file, FILE * fgf_file, int len,
                              int height, int interlace);
void error (char * string);
void message (char * string);
void flush_data_block ();
unsigned long zip_deflate (void);

/* These two are for the deflate modules */
int verbose = 0;
int level = 6;

#ifdef NEED_STCALLS
extern int fflush (FILE *);
extern long unsigned int fwrite (const void *, long unsigned int, 
                                 long unsigned int, FILE *);
extern long unsigned int fread (void *, long unsigned int, long unsigned int,
                                FILE *);
extern int printf (const char *, ...);
extern int fputc (int, FILE *);
#endif

FILE * gif_file, * fgf_file;
unsigned char lzw_min_code_size;


void error (char * string)
{
    printf ("giftofgf: %s\n", string);
    exit (1);
}


void message (char * string)
{
    printf ("giftofgf: %s\n", string);
    fflush (stdout);

    return;
}


int main (int argc, char * argv [])
{
    char * usage = "usage: giftofgf giffile [-o outfile]";
    char * out_file;
    char * p;


    /*
     * Parse command-line
     */
    if (argc != 2 && argc != 4) {
        printf ("%s\n", usage);
        exit (1);
    }

    if (argc == 4)
        out_file = argv [3];
    else {
        if (argv [1][0] == '-') {
            printf ("error: bad argument %s.\n", argv [1]);
            printf ("%s\n", usage);
            exit (1);
        }

        out_file = (char *) malloc (strlen (argv [1]) + 3 + 1);
        strcpy (out_file, argv [1]);

        if ((p = strrchr (out_file, '.')) != NULL)
            *p = '\0';

        strcat (out_file, ".fgf");
    }
    
    /*
     * Open GIF file and create FGF file
     */
    if ((gif_file = fopen (argv [1], "r")) == NULL) {
        printf ("error: cannot open %s.\n", argv [1]);
        exit (1);
    }

    if ((fgf_file = fopen (out_file, "w+")) == NULL) {
        printf ("error: cannot open %s.\n", out_file);
        exit (1);
    }

    /*
     * Convert the file
     */
    ConvertGIF (gif_file, fgf_file);

    return (0);
}


static void ConvertGIF (FILE * gif_file, FILE * fgf_file)
{
    unsigned char   buf [16];
    unsigned char   c;
    unsigned char   localColorMap [3][MAXCOLORMAPSIZE];
    int             useGlobalColormap;
    int             bitPixel;
    int             imageCount = 0;
    char            version [4];


    if (! ReadOK (gif_file, buf, 6))
        error ("error reading magic number");

    if (strncmp ((char *) buf, "GIF", 3) != 0)
        error ("not a GIF file");

    strncpy (version, (char *) buf + 3, 3);
    version [3] = '\0';

    if ((strcmp (version, "87a") != 0) && (strcmp (version, "89a") != 0))
        error ("bad version number, not '87a' or '89a'");

    strcpy (buf, "FGF95a");
    if (! WriteOK (fgf_file, buf, 6))
        error ("failed to write header");

    if (! ReadOK (gif_file, buf, 7))
        error ("failed to read screen descriptor");

    GifScreen.Width           = LM_to_uint(buf[0],buf[1]);
    GifScreen.Height          = LM_to_uint(buf[2],buf[3]);
    GifScreen.BitPixel        = 2<<(buf[4]&0x07);
    GifScreen.ColorResolution = (((buf[4]&0x70)>>3)+1);
    GifScreen.Background      = buf[5];
    GifScreen.AspectRatio     = buf[6];

    if (! WriteOK (fgf_file, buf, 7))
        error ("failed to write screen descriptor");

    if (BitSet(buf[4], LOCALCOLORMAP)) {    /* Global Colormap */
        if (ReadColorMap (gif_file, GifScreen.BitPixel, GifScreen.ColorMap))
            error ("error reading global colormap");

        if (WriteColorMap (fgf_file, GifScreen.BitPixel, GifScreen.ColorMap))
            error ("error writing global colormap");
    }


    if (GifScreen.AspectRatio != 0 && GifScreen.AspectRatio != 49) {
        float   r;
        r = ( (float) GifScreen.AspectRatio + 15.0 ) / 64.0;
#if 0
        message("warning - non-square pixels; to fix do a "
                "'pnmscale -%cscale %g'",
                r < 1.0 ? 'x' : 'y',
                r < 1.0 ? 1.0 / r : r );
#endif
    }

    for (;;) {
        if (! ReadOK (gif_file, &c, 1))
            error ("EOF / read error on image data");

        if (! WriteOK (fgf_file, &c, 1))
            error ("write error on descriptor type");

        if (c == ';') {         /* GIF terminator */
            return;
        }

        if (c == '!') {         /* Extension */
            if (! ReadOK (gif_file, &c, 1))
                error ("OF / read error on extension function code");
            if (! WriteOK (fgf_file, &c, 1))
                error ("write error on descriptor type");

            DoExtension(gif_file, c);
            continue;
        }

        if (c != ',') {         /* Not a valid start character */
            message ("bogus character 0x/*%02x*/, ignoring");/*, (int) c);*/
            continue;
        }

        ++imageCount;

        if (! ReadOK (gif_file, buf, 9))
            error ("couldn't read left/top/width/height");
        if (! WriteOK (fgf_file, buf, 9))
            error ("couldn't write left/top/width/height");

        useGlobalColormap = ! BitSet(buf[8], LOCALCOLORMAP);

        bitPixel = 1<<((buf[8]&0x07)+1);

        if (! useGlobalColormap) {
            if (ReadColorMap (gif_file, bitPixel, localColorMap))
                error ("error reading local colormap");
            if (WriteColorMap (fgf_file, bitPixel, localColorMap))
                error ("error writing local colormap");

            ConvertImageData (gif_file, fgf_file, LM_to_uint(buf[4],buf[5]),
                              LM_to_uint(buf[6],buf[7]),
                              BitSet(buf[8], INTERLACE));
        } else {
            ConvertImageData (gif_file, fgf_file, LM_to_uint(buf[4],buf[5]),
                              LM_to_uint(buf[6],buf[7]),
                              BitSet(buf[8], INTERLACE));
        }
    }
}

static int ReadColorMap (fd, number, buffer)
FILE           *fd;
int            number;
unsigned char  buffer[3][MAXCOLORMAPSIZE];
{
       int             i;
       unsigned char   rgb[3];


       for (i = 0; i < number; ++i) {
               if (! ReadOK(fd, rgb, sizeof(rgb)))
                   error ("bad colormap");

               buffer[CM_RED][i] = rgb[0] ;
               buffer[CM_GREEN][i] = rgb[1] ;
               buffer[CM_BLUE][i] = rgb[2] ;
       }

       return FALSE;
}


static int WriteColorMap (FILE * fd, int number,
                          unsigned char buffer [3][MAXCOLORMAPSIZE])
{
    int             i;
    unsigned char   rgb[3];


    for (i = 0; i < number; ++i) {
        rgb [0] = buffer [CM_RED][i];
        rgb [1] = buffer [CM_GREEN][i];
        rgb [2] = buffer [CM_BLUE][i];
        
        if (! WriteOK(fd, rgb, sizeof (rgb)))
            error ("error writing colormap" );
    }

    return FALSE;
}


static int
DoExtension(fd, label)
FILE   *fd;
int    label;
{
       static char     buf[256];
       char            *str;

       switch (label) {
       case 0x01:              /* Plain Text Extension */
               str = "Plain Text Extension";
#ifdef notdef
               if (GetAndWriteDataBlock (fd, (unsigned char*) buf) == 0)
                       ;

               lpos   = LM_to_uint(buf[0], buf[1]);
               tpos   = LM_to_uint(buf[2], buf[3]);
               width  = LM_to_uint(buf[4], buf[5]);
               height = LM_to_uint(buf[6], buf[7]);
               cellw  = buf[8];
               cellh  = buf[9];
               foreground = buf[10];
               background = buf[11];

               while (GetAndWriteDataBlock (fd, (unsigned char*) buf) != 0) {
                       PPM_ASSIGN(image[ypos][xpos],
                                       cmap[CM_RED][v],
                                       cmap[CM_GREEN][v],
                                       cmap[CM_BLUE][v]);
                       ++index;
               }

               return FALSE;
#else
               break;
#endif
       case 0xff:              /* Application Extension */
               str = "Application Extension";
               break;
       case 0xfe:              /* Comment Extension */
               str = "Comment Extension";
               while (GetAndWriteDataBlock (fd, (unsigned char*) buf) != 0)
                   ;
               return FALSE;
       case 0xf9:              /* Graphic Control Extension */
               str = "Graphic Control Extension";
               (void) GetAndWriteDataBlock (fd, (unsigned char*) buf);
               Gif89.disposal    = (buf[0] >> 2) & 0x7;
               Gif89.inputFlag   = (buf[0] >> 1) & 0x1;
               Gif89.delayTime   = LM_to_uint(buf[1],buf[2]);
               if ((buf[0] & 0x1) != 0)
                       Gif89.transparent = buf[3];

               while (GetAndWriteDataBlock (fd, (unsigned char*) buf) != 0)
                       ;
               return FALSE;
       default:
               str = buf;
               sprintf(buf, "UNKNOWN (0x%02x)", label);
               break;
       }

       while (GetAndWriteDataBlock (fd, (unsigned char*) buf) != 0)
               ;

       return FALSE;
}

int    ZeroDataBlock = FALSE;

static int
GetDataBlock(fd, buf)
FILE           *fd;
unsigned char  *buf;
{
       unsigned char   count;

       if (! ReadOK (fd, &count, 1)) {
               message("error in getting DataBlock size" );
               return -1;
       }

       ZeroDataBlock = count == 0;

       if ((count != 0) && (! ReadOK (fd, buf, count))) {
               message("error in reading DataBlock" );
               return -1;
       }

       return count;
}


static int GetAndWriteDataBlock (FILE * fd, unsigned char * buf)
{
       unsigned char   count;

       if (! ReadOK (gif_file, &count, 1)) {
               message("error in getting DataBlock size" );
               return -1;
       }
       if (! WriteOK (fgf_file, &count, 1)) {
           error ("error in writing DataBlock size");
           return -1;
       }

       if ((count != 0) && (! ReadOK (gif_file, buf, count))) {
               message("error in reading DataBlock" );
               return -1;
       }
       if ((count != 0) && (! WriteOK (fgf_file, buf, count))) {
           error ("error in writing DataBlock");
           return -1;
       }

       return count;
}


static int
GetCode(fd, code_size, flag)
FILE   *fd;
int    code_size;
int    flag;
{
       static unsigned char    buf[280];
       static int              curbit, lastbit, done, last_byte;
       int                     i, j, ret;
       unsigned char           count;

       if (flag) {
               curbit = 0;
               lastbit = 0;
               done = FALSE;
               return 0;
       }

       if ( (curbit+code_size) >= lastbit) {
               if (done) {
                       if (curbit >= lastbit)
                               error("ran off the end of my bits" );
                       return -1;
               }
               buf[0] = buf[last_byte-2];
               buf[1] = buf[last_byte-1];

               if ((count = GetDataBlock(fd, &buf[2])) == 0)
                       done = TRUE;

               last_byte = 2 + count;
               curbit = (curbit - lastbit) + 16;
               lastbit = (2+count)*8 ;
       }

       ret = 0;
       for (i = curbit, j = 0; j < code_size; ++i, ++j)
               ret |= ((buf[ i / 8 ] & (1 << (i % 8))) != 0) << j;

       curbit += code_size;

       return ret;
}

static int
LZWReadByte(fd, flag, input_code_size)
FILE   *fd;
int    flag;
int    input_code_size;
{
       static int      fresh = FALSE;
       int             code, incode;
       static int      code_size, set_code_size;
       static int      max_code, max_code_size;
       static int      firstcode, oldcode;
       static int      clear_code, end_code;
       static int      table[2][(1<< MAX_LZW_BITS)];
       static int      stack[(1<<(MAX_LZW_BITS))*2], *sp;
       register int    i;

       if (flag) {
               set_code_size = input_code_size;
               code_size = set_code_size+1;
               clear_code = 1 << set_code_size ;
               end_code = clear_code + 1;
               max_code_size = 2*clear_code;
               max_code = clear_code+2;

               GetCode(fd, 0, TRUE);
               
               fresh = TRUE;

               for (i = 0; i < clear_code; ++i) {
                       table[0][i] = 0;
                       table[1][i] = i;
               }
               for (; i < (1<<MAX_LZW_BITS); ++i)
                       table[0][i] = table[1][0] = 0;

               sp = stack;

               return 0;
       } else if (fresh) {
               fresh = FALSE;
               do {
                       firstcode = oldcode =
                               GetCode(fd, code_size, FALSE);
               } while (firstcode == clear_code);
               return firstcode;
       }

       if (sp > stack)
               return *--sp;

       while ((code = GetCode(fd, code_size, FALSE)) >= 0) {
               if (code == clear_code) {
                       for (i = 0; i < clear_code; ++i) {
                               table[0][i] = 0;
                               table[1][i] = i;
                       }
                       for (; i < (1<<MAX_LZW_BITS); ++i)
                               table[0][i] = table[1][i] = 0;
                       code_size = set_code_size+1;
                       max_code_size = 2*clear_code;
                       max_code = clear_code+2;
                       sp = stack;
                       firstcode = oldcode =
                                       GetCode(fd, code_size, FALSE);
                       return firstcode;
               } else if (code == end_code) {
                       int             count;
                       unsigned char   buf[260];

                       if (ZeroDataBlock)
                               return -2;

                       while ((count = GetDataBlock(fd, buf)) > 0)
                               ;

                       if (count != 0)
                               message("missing EOD in data stream (common occurence)");
                       return -2;
               }

               incode = code;

               if (code >= max_code) {
                       *sp++ = firstcode;
                       code = oldcode;
               }

               while (code >= clear_code) {
                       *sp++ = table[1][code];
                       if (code == table[0][code])
                               error("circular table entry BIG ERROR");
                       code = table[0][code];
               }

               *sp++ = firstcode = table[1][code];

               if ((code = max_code) <(1<<MAX_LZW_BITS)) {
                       table[0][code] = oldcode;
                       table[1][code] = firstcode;
                       ++max_code;
                       if ((max_code >= max_code_size) &&
                               (max_code_size < (1<<MAX_LZW_BITS))) {
                               max_code_size *= 2;
                               ++code_size;
                       }
               }

               oldcode = incode;

               if (sp > stack)
                       return *--sp;
       }
       return code;
}


static struct {
    int  cur_size;
    char data_block [256];

    int  done;
} g;

static void ConvertImageData (FILE * gif_file, FILE * fgf_file, int len,
                              int height, int interlace)
{
    g.cur_size = 0;
    g.done     = 0;

    /*
     *  Initialize the LZW compression routine
     */
    if (! ReadOK (gif_file, &lzw_min_code_size, 1))
        error("EOF / read error on image data");

    if (LZWReadByte (gif_file, TRUE, lzw_min_code_size) < 0)
        error("error reading image");

    /*
     * Run compressor against this image data;
     * routines file_read and file_write below, are called by deflate
     */
    zip_deflate ();

    /*
     * Flush last data block (if needed) and write out block terminator
     */
    flush_data_block ();

    return;
}


unsigned long zip_deflate (void)
{
    extern void bi_init (FILE *);
    extern void ct_init (unsigned short * attr, int * method);
    extern void lm_init (int pack_level, unsigned short * flags);
    extern unsigned long deflate (void);

    unsigned short attr = 0;  /* BINARY  */
    int method = 8;           /* DEFLATE */
    unsigned short flags = 8;
    FILE * the_file = (FILE *) 1;


    bi_init (the_file);       /* dummy value */
    ct_init (&attr, &method);
    lm_init (6, &flags);

    return (deflate ());
}


int file_read (char * buf, unsigned size)
{
    int length = 0;
    int v;


    while (length < size && !g.done &&
           (v = LZWReadByte (gif_file, FALSE, lzw_min_code_size)) >= 0)
        buf [length++] = v;

    if (v < 0)
        g.done = 1;

    return (length);
}



int file_write (char * buf, int size, int nitems, FILE * zfile)
{
    int bytes_written = 0;


    while (bytes_written < nitems) {
        g.data_block [g.cur_size++] = buf [bytes_written++];

        if (g.cur_size == 254) {
            fputc (g.cur_size & 0xff, fgf_file);

            if (! WriteOK (fgf_file, g.data_block, 254))
                error ("error writing data block");

            g.cur_size = 0;
        }
    }

    return (bytes_written);
}


void flush_data_block ()
{
    /*
     * Write out residual pixel data
     */
    if (g.cur_size > 0) {
        fputc (g.cur_size & 0xff, fgf_file);

        if (! WriteOK (fgf_file, g.data_block, g.cur_size))
            error ("error writing data block");
    }

    /*
     * Write out block terminator
     */
    g.cur_size = 0;
    if (! WriteOK (fgf_file, &g.cur_size, 1))
        error ("error writing block terminator");

    return;
}
