/*
 * ppmtofgf.c -- Convert a portable pixmap to a FGF file
 *
 * Jeremy Wohl - January 8, 1995
 *
 * Portions of this file are derived from the ppmtogif.c module by Jeff
 * Poskanzer, as included in the NetPBM package (release 7 December 1993).
 *
 * The deflate routines were written by Mark Adler et al. and
 * originate in the zip package distributed by the Info-Zip group.
 *
 * This software is provided "as is" without express or implied warranty.
 *
 */

#include "ppm.h"
#include "ppmcmap.h"

extern unsigned long deflate ();

#define MAXCOLORS 256

/*
 * Pointer to function returning an int
 */
typedef int (* ifunptr) ARGS((int, int, int));

static int colorstobpp ARGS(( int colors ));
static int GetPixel ARGS(( int truecolor, int x, int y ));
static void BumpPixel ARGS(( void ));
static int FGFNextPixel ARGS(( ifunptr getpixel, int truecolor ));
static void FGFEncode ARGS(( FILE* fp, int GWidth, int GHeight, int GInterlace, int Truecolor, int Background, int Transparent, int BitsPerPixel, int* Red, int* Green, int* Blue, ifunptr GetPixel ));
int Red[MAXCOLORS],Green[MAXCOLORS],Blue[MAXCOLORS],perm[MAXCOLORS],permi[MAXCOLORS];
int colors;
pixval maxtmp;

static void Putword ARGS(( int w, FILE* fp ));
static int sqr ARGS((int x));
static int closestcolor ARGS((pixel color));

static pixel** pixels;
static colorhash_table cht;

void flush_data_block ();
void error ARGS((char * string));
unsigned long zip_deflate ARGS((void));

/* These two are for the deflate modules */
int verbose = 0;
int level = 6;

#ifdef NEED_STCALLS
extern int fflush (FILE *);
extern long unsigned int fwrite (const void *, long unsigned int, 
                                 long unsigned int, FILE *);
extern long unsigned int fread (void *, long unsigned int, long unsigned int,
                                FILE *);
extern int printf (const char *, ...);
extern int fputc (int, FILE *);
extern int fclose (FILE *);
#endif

void error (char * string)
{
    pm_error ("ppmtofgf: %s", string);
}


int
main( argc, argv )
    int argc;
    char* argv[];
    {
    FILE* ifp;
    int argn, rows, cols, i,j,k, BitsPerPixel;
    int interlace, truecolor, sort, map, transparent;
    pixel transcolor;
    char *mapfile;
    pixval maxval;
    colorhist_vector chv;
    char* usage = "[-interlace] [[-truecolor] | [-sort] [-map mapfile]] "
        "[-transparent color] [ppmfile]";


    ppm_init( &argc, argv );

    argn = 1;
    interlace = 0;
    truecolor = 0;
    sort = 0;
    map = 0;
    transparent = -1;

    while ( argn < argc && argv[argn][0] == '-' && argv[argn][1] != '\0' )
        {
        if ( pm_keymatch( argv[argn], "-interlace", 2 ) )
            interlace = 1;
        else if (pm_keymatch (argv [argn], "-truecolor", 4))
            truecolor = 1;
        else if ( pm_keymatch( argv[argn], "-nointerlace", 2 ) )
            interlace = 0;
        else if ( pm_keymatch( argv[argn], "-sort", 2 ) ) {
            if (truecolor)
                pm_usage (usage);
            sort = 1;
        }
        else if ( pm_keymatch( argv[argn], "-map", 2 ) ) {
            if (truecolor)
                pm_usage (usage);
            map = 1;
            if (++argn < argc)
              mapfile = argv[argn];
            else pm_usage(usage); }
        else if ( pm_keymatch( argv[argn], "-transparent", 4 ) ) {
            transparent = 1;
            if (++argn < argc)
                transcolor = ppm_parsecolor( argv[argn], 255 );
            else
                pm_usage(usage);
        }
        else
            pm_usage( usage );
        ++argn;
        }

    /* Read the colormap from another file. */
    if (map) {
      ifp = pm_openr(mapfile);
      pixels = ppm_readppm( ifp, &cols, &rows, &maxval );
      pm_close( ifp );

      /* Figure out the colormap from the <mapfile>. */
      pm_message( "computing other colormap..." );
      chv = ppm_computecolorhist( pixels, cols, rows, MAXCOLORS, &colors );

      ppm_freearray(pixels,rows); }

    if ( argn < argc )
        {
        ifp = pm_openr( argv[argn] );
        ++argn;
        }
    else
        ifp = stdin;

    if ( argn != argc )
        pm_usage( usage );

    pixels = ppm_readppm( ifp, &cols, &rows, &maxtmp );
    if (!map)
      maxval=maxtmp;

    pm_close( ifp );

    if (truecolor) {
        transparent = PPM_GETR(transcolor) << 16 | PPM_GETG(transcolor) << 8 |
            PPM_GETB(transcolor);
    }
    else {
        /* Figure out the colormap. */
        if (!map) {
            pm_message( "computing colormap..." );
            chv = ppm_computecolorhist( pixels, cols, rows, MAXCOLORS,
                                        &colors ); }

        if ( chv == (colorhist_vector) 0 )
            pm_error(
                "too many colors - try doing a 'ppmquant %d'"
                " or specify '-truecolor'", MAXCOLORS );
        pm_message( "%d colors found", colors );

        /* Now turn the ppm colormap into the appropriate FGF colormap. */
        if ( maxval > 255 )
            pm_message(
                "maxval is not 255 - automatically rescaling colors" );
        for ( i = 0; i < colors; ++i )
            {
                if ( maxval == 255 )
                    {
                        Red[i] = PPM_GETR( chv[i].color );
                        Green[i] = PPM_GETG( chv[i].color );
                        Blue[i] = PPM_GETB( chv[i].color );
                    }
                else
                    {
                        Red[i] = (int) PPM_GETR( chv[i].color ) *
                            255 / maxval;
                        Green[i] = (int) PPM_GETG( chv[i].color ) *
                            255 / maxval;
                        Blue[i] = (int) PPM_GETB( chv[i].color ) *
                            255 / maxval;
                    }
            }

        /* Sort the colormap */
        for (i=0;i<colors;i++)
            permi[i]=i;
        if (sort) {
            pm_message("sorting colormap");
            for (i=0;i<colors;i++)
                for (j=i+1;j<colors;j++)
                    if (((Red[i]*MAXCOLORS)+Green[i])*MAXCOLORS+Blue[i] >
                        ((Red[j]*MAXCOLORS)+Green[j])*MAXCOLORS+Blue[j]) {
                        k=permi[i]; permi[i]=permi[j]; permi[j]=k;
                        k=Red[i]; Red[i]=Red[j]; Red[j]=k;
                        k=Green[i]; Green[i]=Green[j]; Green[j]=k;
                        k=Blue[i]; Blue[i]=Blue[j]; Blue[j]=k; } }
        for (i=0;i<colors;i++)
            perm[permi[i]]=i;

        BitsPerPixel = colorstobpp( colors );

        /* And make a hash table for fast lookup. */
        cht = ppm_colorhisttocolorhash( chv, colors );
        ppm_freecolorhist( chv );

        /* figure out the transparent colour index */
        if (transparent > 0) {
            transparent = ppm_lookupcolor( cht, &transcolor );
            if (transparent == -1)
                transparent = closestcolor( transcolor );
            else
                transparent = perm[transparent];
        }
    }

    /* All set, let's do it. */
    FGFEncode(
        stdout, cols, rows, interlace, truecolor, 0, transparent, BitsPerPixel,
        Red, Green, Blue, GetPixel );

    exit( 0 );
    }

static int
colorstobpp( colors )
int colors;
    {
    int bpp;

    if ( colors <= 2 )
        bpp = 1;
    else if ( colors <= 4 )
        bpp = 2;
    else if ( colors <= 8 )
        bpp = 3;
    else if ( colors <= 16 )
        bpp = 4;
    else if ( colors <= 32 )
        bpp = 5;
    else if ( colors <= 64 )
        bpp = 6;
    else if ( colors <= 128 )
        bpp = 7;
    else if ( colors <= 256 )
        bpp = 8;
    else
        pm_error( "can't happen" );

    return bpp;
    }


static int
sqr(x)
int x;
  {
  return x*x;
  }


static int
closestcolor(color)
pixel color;
  {
  int i,r,g,b,d,
      imin,dmin;

  r=(int)PPM_GETR(color)*255/maxtmp;
  g=(int)PPM_GETG(color)*255/maxtmp;
  b=(int)PPM_GETB(color)*255/maxtmp;

  dmin=1000000;
  for (i=0;i<colors;i++) {
    d=sqr(r-Red[i])+sqr(g-Green[i])+sqr(b-Blue[i]);
    if (d<dmin) {
      dmin=d;
      imin=i; } }
  ppm_addtocolorhash(cht,&color,permi[imin]);
  return imin;
  }


static int
GetPixel( truecolor, x, y )
int truecolor, x, y;
{
    int color;


    if (truecolor) {
        color = (PPM_GETR(pixels[y][x]) << 16) |
            (PPM_GETG(pixels[y][x]) << 8) | PPM_GETB(pixels[y][x]);
    }
    else {
        color = ppm_lookupcolor( cht, &pixels[y][x] );
        if (color == -1)
            color = closestcolor(pixels[y][x]);
        else
            color=perm[color];
    }

    return color;
}


static int Width, Height;
static int curx, cury;
static long CountDown;
static int Pass = 0;
static int Interlace;

/*
 * Bump the 'curx' and 'cury' to point to the next pixel
 */
static void
BumpPixel()
{
        /*
         * Bump the current X position
         */
        ++curx;

        /*
         * If we are at the end of a scan line, set curx back to the beginning
         * If we are interlaced, bump the cury to the appropriate spot,
         * otherwise, just increment it.
         */
        if( curx == Width ) {
                curx = 0;

                if( !Interlace )
                        ++cury;
                else {
                     switch( Pass ) {

                       case 0:
                          cury += 8;
                          if( cury >= Height ) {
                                ++Pass;
                                cury = 4;
                          }
                          break;

                       case 1:
                          cury += 8;
                          if( cury >= Height ) {
                                ++Pass;
                                cury = 2;
                          }
                          break;

                       case 2:
                          cury += 4;
                          if( cury >= Height ) {
                             ++Pass;
                             cury = 1;
                          }
                          break;

                       case 3:
                          cury += 2;
                          break;
                        }
                }
        }
}


/*
 * Write out a word to the FGF file
 */
static void
Putword( w, fp )
int w;
FILE* fp;
{
        fputc( w & 0xff, fp );
        fputc( (w / 256) & 0xff, fp );
}


/*
 * Return the next pixel from the image
 */
static int
FGFNextPixel( getpixel, truecolor )
ifunptr getpixel;
int truecolor;
{
        int r;

        if( CountDown == 0 )
                return EOF;

        --CountDown;

        r = ( * getpixel ) (truecolor, curx, cury);

        BumpPixel();

        return r;
}


static struct {
    ifunptr        getpixel;
    int            truecolor;
    FILE         * fgf_file;
    unsigned char  rgb_buffer [3];
    int            rgb_buffer_count;

    /* used by file_write */
    int            cur_size;
    unsigned char  data_block [256];
} g;

static void
FGFEncode( fp, GWidth, GHeight, GInterlace, Truecolor, Background,
           Transparent, BitsPerPixel, Red, Green, Blue, GetPixel )

FILE* fp;
int GWidth, GHeight;
int GInterlace;
int Truecolor;
int Background;
int Transparent;
int BitsPerPixel;
int Red[], Green[], Blue[];
ifunptr GetPixel;
{
        int B = 0;
        int RWidth, RHeight;
        int LeftOfs, TopOfs;
        int Resolution;
        int ColorMapSize;
        int InitCodeSize;
        int i;


        g.getpixel = GetPixel;
        g.truecolor = Truecolor;
        g.fgf_file = fp;

        g.rgb_buffer_count = 0;

        g.cur_size = 0;

        Interlace = GInterlace;

        ColorMapSize = 1 << BitsPerPixel;

        RWidth = Width = GWidth;
        RHeight = Height = GHeight;
        LeftOfs = TopOfs = 0;

        Resolution = BitsPerPixel;

        /*
         * Calculate number of bits we are expecting
         */
        CountDown = (long)Width * (long)Height;

        /*
         * Indicate which pass we are on (if interlace)
         */
        Pass = 0;

        /*
         * The initial code size
         */
        if (!Truecolor) {
            if( BitsPerPixel <= 1 )
                InitCodeSize = 2;
            else
                InitCodeSize = BitsPerPixel;
        }

        /*
         * Set up the current x and y position
         */
        curx = cury = 0;

        /*
         * Write the Magic header
         */
        fwrite("FGF95a", 1, 6, fp);

        /*
         * Write out the screen width and height
         */
        Putword( RWidth, fp );
        Putword( RHeight, fp );
        
        if (Truecolor)
            B = 0;
        else {
            /*
             * Indicate that there is a global colour map
             */
            B = 0x80;       /* Yes, there is a color map */

            /*
             * OR in the resolution
             */
            B |= ((Resolution - 1) << 5);

            /*
             * OR in the Bits per Pixel
             */
            B |= (BitsPerPixel - 1);
        }

        /*
         * Write it out
         */
        fputc( B, fp );

        /*
         * Write out the Background colour
         */
        fputc( Background, fp );

        /*
         * Byte of 0's (future expansion)
         */
        fputc( 0, fp );

        if (!Truecolor) {
            /*
             * Write out the Global Colour Map
             */
            for( i=0; i<ColorMapSize; ++i ) {
                fputc( Red[i], fp );
                fputc( Green[i], fp );
                fputc( Blue[i], fp );
            }
        }

        if (Truecolor) {
            /*
             * Write out truecolor extension
             */
            fputc (0x21, fp);
            fputc (0x02, fp);
            fputc (7, fp);
            fputc ((Transparent >= 0) ? 0x01 : 0, fp);
            fputc ((Transparent >> 16) & 0xff, fp);
            fputc ((Transparent >> 8)  & 0xff, fp);
            fputc (Transparent & 0xff, fp);
            fputc (0, fp);
            fputc (0, fp);
            fputc (0, fp);
            fputc (0, fp);
        }
        else {
            /*
             * Write out extension for transparent colour index, if necessary.
             */
            if ( Transparent >= 0 ) {
                fputc( '!', fp );
                fputc( 0xf9, fp );
                fputc( 4, fp );
                fputc( 1, fp );
                fputc( 0, fp );
                fputc( 0, fp );
                fputc( Transparent, fp );
                fputc( 0, fp );
            }
        }

        /*
         * Write an Image separator
         */
        fputc( ',', fp );

        /*
         * Write the Image header
         */

        Putword( LeftOfs, fp );
        Putword( TopOfs, fp );
        Putword( Width, fp );
        Putword( Height, fp );

        /*
         * Write out whether or not the image is interlaced
         */
        if( Interlace )
                fputc( 0x40, fp );
        else
                fputc( 0x00, fp );

        /*
         * Initialize and run compressor against this image data;
         * routines file_read and file_write below, are called by deflate
         */
        zip_deflate ();

        /*
         * Flush last data block (if needed) and write out block terminator
         */
        flush_data_block ();

        /*
         * Write out a Zero-length packet (to end the series)
         */
        fputc( 0, fp );

        /*
         * Write the FGF file terminator
         */
        fputc( ';', fp );

        /*
         * And close the file
         */
        fclose( fp );
}


unsigned long zip_deflate (void)
{
    extern void bi_init (FILE *);
    extern void ct_init (unsigned short * attr, int * method);
    extern void lm_init (int pack_level, unsigned short * flags);
    extern unsigned long deflate (void);

    unsigned short attr = 0;  /* BINARY  */
    int method = 8;           /* DEFLATE */
    unsigned short flags = 8;
    FILE * the_file = (FILE *) 1;


    bi_init (the_file);       /* dummy value */
    ct_init (&attr, &method);
    lm_init (6, &flags);

    return (deflate ());
}


int file_read (char * buf, unsigned size)
{
    static int done = 0;
    int length = 0;
    int i, v;


    while (length < size && !done) {
        if (g.rgb_buffer_count == 0)
            if ((v = FGFNextPixel (g.getpixel, g.truecolor)) < 0)
                break;
        
        if (g.truecolor) {
            if (g.rgb_buffer_count == 0) {
                g.rgb_buffer_count = 3;

                g.rgb_buffer [0] = (v >> 16) & 0xff;
                g.rgb_buffer [1] = (v >> 8)  & 0xff;
                g.rgb_buffer [2] = v & 0xff;
            }
            
            for (i = 3 - g.rgb_buffer_count;
                 length < size && g.rgb_buffer_count > 0;
                 i++, g.rgb_buffer_count--) {
                buf [length++] = g.rgb_buffer [i];
            }
            
        }
        else
            buf [length++] = (unsigned char) v;
    }

    if (v < 0)
        done = 1;

    return (length);
}


int file_write (char * buf, int size, int nitems, FILE * zfile)
{
    int bytes_written = 0;


    while (bytes_written < nitems) {
        g.data_block [g.cur_size++] = buf [bytes_written++];

        if (g.cur_size == 254) {
            fputc (g.cur_size & 0xff, g.fgf_file);

            if (fwrite (g.data_block, 254, 1, g.fgf_file) == 0)
                pm_error ("error writing data block");

            g.cur_size = 0;
        }
    }

    return (bytes_written);
}


void flush_data_block ()
{
    /*
     * Write out residual pixel data
     */
    if (g.cur_size > 0) {
        fputc (g.cur_size & 0xff, g.fgf_file);

        if (fwrite (g.data_block, g.cur_size, 1, g.fgf_file) == 0)
            pm_error ("error writing data block");
    }

    return;
}
